/*
Copyright 2021.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package trait

import (
	"context"
	"crypto/rand"
	"errors"
	"fmt"

	"github.com/fluxcd/pkg/runtime/conditions"
	"github.com/minio/madmin-go"
	"github.com/minio/minio-go/v7"

	"github.com/minio/minio-go/v7/pkg/credentials"
	"k8s.io/apimachinery/pkg/runtime"
	ctrl "sigs.k8s.io/controller-runtime"
	"sigs.k8s.io/controller-runtime/pkg/client"
	"sigs.k8s.io/controller-runtime/pkg/log"

	corev1 "k8s.io/api/core/v1"
	traitv1alpha1 "libre.sh/controller/apis/trait/v1alpha1"
	"libre.sh/controller/pkg/utils"
)

// BucketReconciler reconciles a Bucket object
type BucketReconciler struct {
	client.Client
	Scheme *runtime.Scheme
}

//+kubebuilder:rbac:groups=trait.libre.sh,resources=buckets,verbs=get;list;watch;create;update;patch;delete
//+kubebuilder:rbac:groups=trait.libre.sh,resources=buckets/status,verbs=get;update;patch
//+kubebuilder:rbac:groups=trait.libre.sh,resources=buckets/finalizers,verbs=update

// Reconcile is part of the main kubernetes reconciliation loop which aims to
// move the current state of the cluster closer to the desired state.
// TODO(user): Modify the Reconcile function to compare the state specified by
// the Bucket object against the actual cluster state, and then
// perform operations to make the cluster state reflect the state specified by
// the user.
//
// For more details, check Reconcile and its Result here:
// - https://pkg.go.dev/sigs.k8s.io/controller-runtime@v0.11.0/pkg/reconcile
func (r *BucketReconciler) Reconcile(ctx context.Context, req ctrl.Request) (ctrl.Result, error) {
	_ = log.FromContext(ctx)

	var bucket traitv1alpha1.Bucket
	err := r.Client.Get(ctx, req.NamespacedName, &bucket)
	if err != nil {
		return ctrl.Result{}, client.IgnoreNotFound(err)
	}
	// bucket.Status = traitv1alpha1.BucketStatus{}
	defer utils.UpdateStatus(ctx, r.Client, &bucket)

	libreContext, err := utils.NewContext(ctx, r.Client, req.Namespace)
	if err != nil {
		return utils.HandleLibreContextError(err, &bucket)
	}
	conditions.Delete(&bucket, traitv1alpha1.LibreContextCondition)

	provider := libreContext.GetS3Provider()
	if provider == nil {
		// TODO better handling
		return ctrl.Result{}, errors.New("provider not found")
	}

	minioClient, err := minio.New(provider.Spec.Endpoint, &minio.Options{
		Creds:  credentials.NewStaticV4(provider.Spec.AccessKey, provider.Spec.SecretKey, ""),
		Secure: !provider.Spec.Insecure,
	})
	if err != nil {
		// TODO better handling
		return ctrl.Result{}, errors.New("failed to connect to s3")
	}
	minioAdminClient, err := madmin.New(provider.Spec.Endpoint, provider.Spec.AccessKey, provider.Spec.SecretKey, !provider.Spec.Insecure)
	if err != nil {
		// TODO better handling
		return ctrl.Result{}, errors.New("failed to connect to admin s3")
	}

	name := fmt.Sprintf("%s-%s", libreContext.Object.Name, req.Name)

	buckerExits, _ := minioClient.BucketExists(ctx, name)
	if !buckerExits {
		err = minioClient.MakeBucket(ctx, name, minio.MakeBucketOptions{Region: provider.Spec.Region})
		if err != nil {
			return ctrl.Result{}, errors.New("failed to create s3 bucket")
		}

	}

	var secretKey string
	user, err := minioAdminClient.GetUserInfo(ctx, name)
	if err != nil {
		key := make([]byte, 32)
		_, err := rand.Read(key)
		if err != nil {
			return ctrl.Result{}, err
		}
		secretKey = string(key)
		err = minioAdminClient.AddUser(ctx, name, secretKey)
		if err != nil {
			return ctrl.Result{}, err
		}

	}
	secretKey = user.SecretKey

	// TODO Policy

	var secret corev1.Secret
	secret.Name = fmt.Sprintf("%s.bucket.trait.libre.sh", req.Name)
	secret.Namespace = fmt.Sprintf("%s.bucket.trait.libre.sh", req.Name)
	_, err = ctrl.CreateOrUpdate(ctx, r.Client, &secret, func() error {
		secret.StringData = map[string]string{
			"accessKey": name,
			"secretKey": secretKey,
		}
		return nil
	})
	if err != nil {
		return ctrl.Result{}, err
	}

	return ctrl.Result{}, nil
}

// SetupWithManager sets up the controller with the Manager.
func (r *BucketReconciler) SetupWithManager(mgr ctrl.Manager) error {
	return ctrl.NewControllerManagedBy(mgr).
		For(&traitv1alpha1.Bucket{}).
		Complete(r)
}
