/*
Copyright 2021.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package trait

import (
	"context"
	"fmt"

	"github.com/fluxcd/pkg/runtime/conditions"
	corev1 "k8s.io/api/core/v1"
	"k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
	"k8s.io/apimachinery/pkg/runtime"
	"k8s.io/apimachinery/pkg/types"
	traitv1alpha1 "libre.sh/controller/apis/trait/v1alpha1"
	"libre.sh/controller/pkg/utils"
	ctrl "sigs.k8s.io/controller-runtime"
	"sigs.k8s.io/controller-runtime/pkg/client"
	"sigs.k8s.io/controller-runtime/pkg/log"
)

// PostgresReconciler reconciles a Postgres object
type PostgresReconciler struct {
	client.Client
	Scheme *runtime.Scheme
}

//+kubebuilder:rbac:groups=trait.libre.sh,resources=postgres,verbs=get;list;watch;create;update;patch;delete
//+kubebuilder:rbac:groups=trait.libre.sh,resources=postgres/status,verbs=get;update;patch
//+kubebuilder:rbac:groups=trait.libre.sh,resources=postgres/finalizers,verbs=update

// Reconcile is part of the main kubernetes reconciliation loop which aims to
// move the current state of the cluster closer to the desired state.
// TODO(user): Modify the Reconcile function to compare the state specified by
// the Postgres object against the actual cluster state, and then
// perform operations to make the cluster state reflect the state specified by
// the user.
//
// For more details, check Reconcile and its Result here:
// - https://pkg.go.dev/sigs.k8s.io/controller-runtime@v0.11.0/pkg/reconcile
func (r *PostgresReconciler) Reconcile(ctx context.Context, req ctrl.Request) (ctrl.Result, error) {
	_ = log.FromContext(ctx)

	var trait traitv1alpha1.Postgres
	err := r.Client.Get(ctx, req.NamespacedName, &trait)
	if err != nil {
		return ctrl.Result{}, client.IgnoreNotFound(err)
	}
	trait.Status = traitv1alpha1.PostgresStatus{}
	defer utils.UpdateStatus(ctx, r.Client, &trait)

	libreContext, err := utils.GetLibreContext(ctx, r.Client, req.Namespace)
	if err != nil {
		return utils.HandleLibreContextError(err, &trait)
	}
	conditions.Delete(&trait, traitv1alpha1.LibreContextCondition)

	var list traitv1alpha1.PostgresList
	err = r.Client.List(ctx, &list, &client.ListOptions{Namespace: req.Namespace})
	if err != nil {
		return ctrl.Result{}, err
	}

	postgres := unstructured.Unstructured{}
	postgres.SetAPIVersion("acid.zalan.do/v1")
	postgres.SetKind("postgresql")
	postgres.SetName(fmt.Sprintf("pg-%s", libreContext.Name))
	postgres.SetNamespace(req.Namespace)
	_, err = ctrl.CreateOrUpdate(ctx, r.Client, &postgres, func() error {
		users := make(map[string][]string)
		databases := make(map[string]string)
		spec, ok := postgres.Object["spec"].(map[string]interface{})
		if ok {
			u, ok := spec["users"].(map[string][]string)
			if ok {
				users = u
			}
			d, ok := spec["databases"].(map[string]string)
			if ok {
				databases = d
			}
		}

		for _, v := range list.Items {
			users[v.Name] = []string{
				"superuser",
				"createdb",
			}
			databases[v.Name] = v.Name
		}
		postgres.Object["spec"] = map[string]interface{}{
			"teamId":            "pg",
			"numberOfInstances": 3,
			"postgresql": map[string]interface{}{
				"version": "11",
			},
			"volume": map[string]interface{}{
				"size": "10G",
			},
			"databases": databases,
			"users":     users,
		}
		return nil
	})
	if err != nil {
		return ctrl.Result{}, err
	}

	if postgres.Object["status"] == nil || postgres.Object["status"].(map[string]interface{})["PostgresClusterStatus"] != "Running" {
		conditions.MarkTrue(&trait, traitv1alpha1.PostgresCondition, "PostgrestNotRunning", "Waiting for Postgres to be running")
		return utils.RetryLater()
	}
	conditions.Delete(&trait, traitv1alpha1.PostgresCondition)

	host := fmt.Sprintf("pg-%s", libreContext.Name)

	secret := corev1.Secret{}
	err = r.Client.Get(ctx, types.NamespacedName{
		Namespace: req.Namespace,
		Name:      fmt.Sprintf("%s.%s.credentials.postgresql.acid.zalan.do", req.Name, host),
	}, &secret)
	if err != nil {
		return ctrl.Result{}, err
	}

	resultSecret := corev1.Secret{}
	resultSecret.Name = fmt.Sprintf("%s.postgres.trait.libre.sh", req.Name)
	resultSecret.Namespace = req.Namespace
	_, err = ctrl.CreateOrUpdate(ctx, r.Client, &resultSecret, func() error {
		resultSecret.StringData = map[string]string{
			"user": req.Name,
			"pass": string(secret.Data["password"]),
			"host": host,
			"name": req.Name,
			"url":  fmt.Sprintf("postgres://%s:%s@%s/%s?sslmode=require", req.Name, secret.Data["password"], host, req.Name),
		}
		return nil
	})
	if err != nil {
		return ctrl.Result{}, err
	}

	utils.MarkReconciliationSucceeded(&trait)
	return ctrl.Result{}, nil
}

// SetupWithManager sets up the controller with the Manager.
func (r *PostgresReconciler) SetupWithManager(mgr ctrl.Manager) error {
	return ctrl.NewControllerManagedBy(mgr).
		For(&traitv1alpha1.Postgres{}).
		Complete(r)
}
