package utils

import (
	"context"
	"errors"
	"time"

	"github.com/fluxcd/pkg/runtime/conditions"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	corev1alpha1 "libre.sh/controller/apis/core/v1alpha1"
	traitv1alpha1 "libre.sh/controller/apis/trait/v1alpha1"
	ctrl "sigs.k8s.io/controller-runtime"
	"sigs.k8s.io/controller-runtime/pkg/client"
	"sigs.k8s.io/controller-runtime/pkg/reconcile"
)

func GetLibreContext(ctx context.Context, c client.Client, namespace string) (corev1alpha1.LibreContext, error) {
	var list corev1alpha1.LibreContextList
	if err := c.List(ctx, &list, &client.ListOptions{
		Namespace: namespace,
	}); err != nil {
		return corev1alpha1.LibreContext{}, err
	}
	if len(list.Items) == 0 {
		return corev1alpha1.LibreContext{}, errors.New("no-libre-context")
	}
	if len(list.Items) > 1 {
		return corev1alpha1.LibreContext{}, errors.New("too-much-libre-context")
	}
	return list.Items[0], nil
}

func HandleLibreContextError(err error, o conditions.Setter) (ctrl.Result, error) {
	if err.Error() == "no-libre-context" {
		conditions.MarkTrue(o, traitv1alpha1.LibreContextCondition, "NoLibreContext", "Waiting for a LibreContext")
		return RetryLater()
	}
	if err.Error() == "too-much-libre-context" {
		conditions.MarkTrue(o, traitv1alpha1.LibreContextCondition, "TooMuchLibreContext", "Waiting for only one LibreContext")
		return RetryLater()
	}
	return reconcile.Result{}, err
}

func RetryLater() (ctrl.Result, error) {
	return ctrl.Result{RequeueAfter: 10 * time.Second}, nil
}

func UpdateStatus(ctx context.Context, c client.Client, o conditions.Setter) func() {
	return func() {
		readyCondition := GetReconciliationSucceeded()
		for _, condition := range o.GetConditions() {
			if condition.Type != "Ready" && condition.Status == metav1.ConditionTrue {
				readyCondition = metav1.Condition{
					Type:    "Ready",
					Status:  metav1.ConditionFalse,
					Reason:  condition.Reason,
					Message: condition.Message,
				}
			}
		}
		conditions.Set(o, &readyCondition)
		c.Status().Update(ctx, o)
	}
}

func GetReconciliationSucceeded() metav1.Condition {
	return metav1.Condition{
		Type:    "Ready",
		Status:  metav1.ConditionTrue,
		Reason:  "ReconciliationSucceeded",
		Message: "Trait reconciliation succeeded",
	}
}

func MarkReconciliationSucceeded(o conditions.Setter) {
	c := GetReconciliationSucceeded()
	conditions.Set(o, &c)
}
